/*+----------------------------------------------------+
  |FILENAME : Mva.c                                    |
  |MVA Version 8.15                                    |
  |Queueing network, Mixed Model Analysis program      |
  |References [1] Quantitative System Performance      |
  |           by Edward D.Lazowska.  Prentice-Hall,Inc.|
  |Copyright (c) 2023 by İݺ All Rights Reserved. |
  +----------------------------------------------------+*/
#define FILE_NAME "Mva.c"

#if defined(PDEF_H)
LIST("# " FILE_NAME " version 8.15")
PROC(Solve,EJ("Solve Queueing Network","׻¹Ԥ륳ޥ"))
INT(TN,     2,EJ("Num of title line","ǡեΥȥιԿ"))
SVEC(Title, 0,TN,EJ("Data file's title","ǡեΥȥ"))
          /*-----------------------------------*
           |  Constant variables               |
           *-----------------------------------*/

FIX(2)
CDEF(NameJT,6,FIX2,EJ("Name of Job type","OPEN CLOSEDΤ줫"))

FIX(3)
CDEF(NameD,6,FIX3,EJ("Name of Discipline","SHARE FCFS DELAYΤ줫"))
          /*-----------------------------------*
           |  Input variables                  |
           *-----------------------------------*/
REAL(Criteria,-311,   EJ("Criteria of MVA iteration","Ϸ̤"))
INT(N,      2,        EJ("Num of nodes","Ρɤο"))
CDEF(Node, 10,N,      EJ("Node Names","Ρɤ̾"))
CVEC(TypeN, 5,N,NameD,EJ("Service discipline","SHAREFCFSDELAY򼨤"))

INT(J,      2,         EJ("Number of Job chains"," ο"))
CDEF(Job,  10,J,       EJ("Job chain Names"," ̾"))
CVEC(TypeJ, 6,J,NameJT,EJ("Job type name OPEN/CLOSED","OPENCLOSED򼨤"))
RVEC(LoadJ,408,J,    EJ("Number of Job or Arrival rate","͸ޤΨ"))

IVEC(L,   2,J,     EJ("Num of nodes on job","֥󤬻Ľο"))
CJAG(NodeL,10,L,Node,EJ("Node names on job chain","ιԤΡ"))
RJAG(VisitL,408,L,   EJ("Visit count","Фˬ"))
RJAG(ServL, 408,L,   EJ("Service time","ˤ"))
          /*-----------------------------------*
           |  Inner variables                  |
           *-----------------------------------*/
INT(Saturate, 8,       EJ("System saturate flag","ƥब˰¤1"))
RMAT(Demand,409,J,N,   EJ("Service demand matrix","ӥ׵"))
RMAT(UtilM, 409,J,N,   EJ("Utilization matrix","ƯΨ򼨤"))
RVEC(UtilN, 409,N,     EJ("Utilization of node","ΡɤФƯΨ"))
RMAT(ResidM,409,J,N,   EJ("Queue length matrix","򼨤"))
RVEC(ResidN,409,N,     EJ("Queue length of node","ΡɤФ"))
RMAT(RespM, 409,J,N,   EJ("Response time matrix","֤򼨤"))
RVEC(RespJ, 409,J,EJ("Response time of job","֥Фα"))
RMAT(VisitM,409,J,N,   EJ("Visit Count matrix","ˬ"))
RMAT(ServM, 409,J,N,   EJ("Service time matrix","ӥֹ"))
RVEC(ThrouJ,407,J,     EJ("Job throuput"," Υ롼ץå"))
          /*-----------------------------------*
           |  Output variables                 |
           *-----------------------------------*/
INT(Count,5,           EJ("Count of iterations","«ޤǤη֤"))
IVEC(S,2,J,          EJ("Num of nodes of spl","ץåȷΥΡɤο"))
CJAG(NodeS,10,S,Node,EJ("Node names of spl","ץåȷΥΡɤ̾"))
RJAG(UtilS, 408,S,     EJ("Utilization of spl","ץåȷβƯΨ"))
RJAG(ResidS,408,S,     EJ("Queue length of spl","ץåȷη"))
RJAG(RespS, 409,S,    EJ("Response time of spl","ץåȷα"))

AUTO("Range LoadJ > 0")
AUTO("Range L > 0")
AUTO("Range VisitL > 0")
AUTO("Range ServL > 0")
AUTO("Range Criteria > 0 <= 100")
AUTO("Criteria")
AUTO("0.1")
AUTO("NameJT")
AUTO("OPEN CLOSED $")
AUTO("NameD")
AUTO("SHARE FCFS DELAY $")


#else
 #if !defined(NO_RAND)
  #define NO_RAND 1
 #endif
 #if defined(LSI_C)
  #define NO_AFTER_VER2 1
 #endif
 #include "../pocom.h"
 CHECK_FILE_NAME

typedef enum { OPEN, CLOSED } openclose;
typedef enum { SHARE, FCFS, DELAY } discipline;

void CheckFCFS(void) {
    Int j,l,n;
    for (j=0;j<J;j++) for (l=0;l<L[j];l++) {
        n=NodeL[j][l];
        if (TypeN[n]==FCFS) {
            if (ServM[j][n]==0.0||ServM[j][n]==ServL[j][l])
            { ServM[j][n]=ServL[j][l]; VisitM[j][n]+=VisitL[j][l]; }
            else Error(Node[n],EJ("Diff serv time","ӥ֤ۤʤ"));
        }
    }
}

void CalcDemandOpenUtilThrou(void) {
    Int j,l,n;
    for (j=0;j<J;j++) for (l=0;l<L[j];l++)
        Demand[j][NodeL[j][l]]+=VisitL[j][l]*ServL[j][l]; /*p.54*/
    for (j=0;j<J;j++) if (TypeJ[j]==OPEN) for (n=0;n<N;n++)
        UtilN[n]+=Demand[j][n]*LoadJ[j]; /*p.138*/
    for (j=0;j<J;j++) if (TypeJ[j]==OPEN) ThrouJ[j]=LoadJ[j];
}

void CheckSaturate(void) {
    Int n; static char buf[60]; size_t pos=0,i; char*p;
    Saturate=false;
    for (n=0;n<N;n++) if (TypeN[n]!=DELAY&&UtilN[n]>=1.00) {
        Saturate=true;
        i=sizeof(buf)-1-pos;
        if (i>71) i=71;
        p=Left__(Node[n],(int)i);
        for (i=0;p[i]&&pos<sizeof buf-1;i++,pos++) buf[pos]=p[i];
        if (pos<sizeof buf-1) buf[pos++]=' '; else break;
    }
    buf[pos]=0;
    if (Saturate) Error(buf,EJ("Node saturated","Ρɤѥ󥯤"));
}

void ModifyClosedDemand(void) {
    Int j,n;
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED)
     for (n=0;n<N;n++) if (TypeN[n]!=DELAY) {
         Demand[j][n]/=1.0f-UtilN[n];    /*p.146 step 2*/
         ServM[j][n] /=1.0f-UtilN[n];    /*p.263 line 1*/
     }
}

void SetInitialResid(void) {
    Int n,j;
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED) for (n=0;n<N;n++)
        ResidM[j][n]=LoadJ[j]/N;
}

real ApproxResid(Int j,Int n) {
    Int k; real Sum=(LoadJ[j]-1)/LoadJ[j]*ResidM[j][n];/*p.144 7.5*/
    for (k=0;k<J;k++) if (TypeJ[k]==CLOSED) if (k!=j) Sum+=ResidM[k][n];
    return Sum;
}

real ApproxSQ(Int j,Int n) {    /* p.263 */
    Int k; real Sum=(LoadJ[j]-1)/LoadJ[j]*ServM[j][n]*ResidM[j][n];
    for (k=0;k<J;k++) if (TypeJ[k]==CLOSED)
        if (k!=j) Sum+=ServM[k][n]*ResidM[k][n];
    return Sum;
}

real SumOfResp(Int j) {
    real Sum=0; Int n;
    for (n=0;n<N;n++) if (TypeJ[j]==CLOSED) Sum+=RespM[j][n];
    return Sum; /*p.139 7.1*/
}

real Iteration(void) {
    Int j,n;  real NewResid,OldResid,ResidErr,MaxResidErr=0.0;
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED) for (n=0;n<N;n++)
     switch (TypeN[n]) {
     case DELAY:RespM[j][n]=Demand[j][n];break; /*p.139 7.3*/
     case SHARE:RespM[j][n]=Demand[j][n]*(1.0f+ApproxResid(j,n));
         break;                                       /*p.144 7.5*/
     case FCFS:RespM[j][n]=VisitM[j][n]*(ServM[j][n]+ApproxSQ(j,n));
         break;                                           /*p.263*/
    }
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED) /*p.139 7.1*/
    { RespJ[j]=SumOfResp(j); ThrouJ[j]=LoadJ[j]/RespJ[j]; }
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED) for (n=0;n<N;n++) {
        OldResid=ResidM[j][n];
        NewResid=ThrouJ[j]*RespM[j][n];             /*p.139 7.2*/
        ResidM[j][n]=NewResid;
        if (NewResid==0.0) NewResid=Criteria;
        if (OldResid==0.0) OldResid=Criteria;
        if (NewResid>=OldResid)
             ResidErr=(NewResid-OldResid)/NewResid;
        else ResidErr=(OldResid-NewResid)/OldResid;
        if (ResidErr>MaxResidErr) MaxResidErr=ResidErr;
    }
    return MaxResidErr*100.0f;
}

void ModifyOpenMeasures(void) {
    Int n,j;
    for (j=0;j<J;j++) if (TypeJ[j]==CLOSED)
     for (n=0;n<N;n++) ResidN[n]+=ResidM[j][n];
    for (j=0;j<J;j++) if (TypeJ[j]==OPEN) for (n=0;n<N;n++) {
     switch (TypeN[n]) {
     case DELAY:RespM[j][n]=Demand[j][n];break;      /*p146 step 3*/
     case SHARE:RespM[j][n]=
                Demand[j][n]*(1.0f+ResidN[n])/(1.0f-UtilN[n]);break;
     case FCFS :RespM[j][n]=
                Demand[j][n]*(1.0f+ResidN[n])/(1.0f-UtilN[n]);break;
     }
     ResidM[j][n]=LoadJ[j]*RespM[j][n];
    }
   for (j=0;j<J;j++) if (TypeJ[j]==OPEN) for (n=0;n<N;n++)
   { ResidN[n]+=ResidM[j][n]; RespJ[j]+=RespM[j][n]; }
}

void CalcPerformanceMeasures(void) {
    Int j,l,n;
    for (n=0;n<N;n++)
    { UtilN[n]=0.0; for (j=0;j<J;j++) Demand[j][n]=0.0; }
    for (j=0;j<J;j++) for (l=0;l<L[j];l++)
        Demand[j][NodeL[j][l]]+=VisitL[j][l]*ServL[j][l]; /*p.54*/
    for (n=0;n<N;n++) for (j=0;j<J;j++)
    { UtilM[j][n]=Demand[j][n]*ThrouJ[j]; UtilN[n]+=UtilM[j][n]; }
}

void Splitting(void) {
    Int j,n,k;
    Alloc("S");
    for (j=0;j<J;j++) for (n=0;n<N;n++) if (Demand[j][n]!=0) S[j]++;
    SizeChangedBy(S,Solve);
    Alloc("NodeS UtilS ResidS RespS");
    for (j=0;j<J;j++) for (k=n=0;n<N;n++) if (Demand[j][n]!=0.0) {
        NodeS[j][k]=n; UtilS[j][k]=UtilM[j][n];
        ResidS[j][k]=ResidM[j][n]; RespS[j][k]=RespM[j][n]; k++;
    }
}

void Solve(void) {
    Check("Node TypeN Job TypeJ LoadJ L NodeL VisitL ServL");
    Alloc("Demand UtilM UtilN ResidM ResidN");
    Alloc("RespM RespJ ThrouJ VisitM ServM");
    CheckFCFS(); CalcDemandOpenUtilThrou(); CheckSaturate();
    ModifyClosedDemand(); SetInitialResid();
    for (Count=0;Iteration()>Criteria;Count++);
    ModifyOpenMeasures(); CalcPerformanceMeasures(); Splitting();
}

#endif
